package envoy.server.processors;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.persistence.EntityExistsException;

import envoy.data.Message;
import envoy.data.Message.MessageStatus;
import envoy.event.MessageStatusChange;
import envoy.server.data.PersistenceManager;
import envoy.server.net.ConnectionManager;
import envoy.server.net.ObjectWriteProxy;
import envoy.util.EnvoyLog;

/**
 * This {@link ObjectProcessor} handles incoming {@link Message}s.
 * <p>
 * Project: <strong>envoy-server-standalone</strong><br>
 * File: <strong>MessageProcessor.java</strong><br>
 * Created: <strong>30.12.2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @since Envoy Server Standalone v0.1-alpha
 */
public class MessageProcessor implements ObjectProcessor<Message> {

	private static final PersistenceManager	persistenceManager	= PersistenceManager.getInstance();
	private static final ConnectionManager	connectionManager	= ConnectionManager.getInstance();
	private static final Logger logger = EnvoyLog.getLogger(MessageProcessor.class);

	@Override
	public void process(Message message, long socketID, ObjectWriteProxy writeProxy) {
		if (message.getStatus() != MessageStatus.WAITING) {
			logger.warning("Received message with invalid status: " + message);
			return;
		}
		message.nextStatus();

		// Convert to server message
		final var serverMessage = new envoy.server.data.Message(message);

		try {

			// Persist the message
			persistenceManager.addMessage(serverMessage);

			// Send the message to the recipient if online
			if (connectionManager.isOnline(message.getRecipientID())) {
				writeProxy.write(connectionManager.getSocketID(message.getRecipientID()), message);

				// Increment status
				message.nextStatus();
				serverMessage.received();
				persistenceManager.updateMessage(serverMessage);

				// Notify the sender about the delivery
				// Note that the exact time stamp might differ slightly
				writeProxy.write(socketID, new MessageStatusChange(message));
			}
		} catch (EntityExistsException e) {
			logger.log(Level.WARNING, "Received " + message + " with an ID that already exists!");
		} catch (IOException e) {
			logger.log(Level.WARNING, "Failed to deliver " + message + ":", e);
		}
	}

	@Override
	public Class<Message> getInputClass() { return Message.class; }
}
