package envoy.server.processors;

import java.io.IOException;
import java.util.logging.Logger;

import envoy.data.User.UserStatus;
import envoy.event.UserStatusChange;
import envoy.server.data.PersistenceManager;
import envoy.server.data.User;
import envoy.server.net.ConnectionManager;
import envoy.server.net.ObjectWriteProxy;
import envoy.util.EnvoyLog;

/**
 * This processor handles incoming {@link UserStatusChange}.<br>
 * <br>
 * Project: <strong>envoy-server-standalone</strong><br>
 * File: <strong>UserStatusChangeProcessor.java</strong><br>
 * Created: <strong>1 Feb 2020</strong><br>
 *
 * @author Leon Hofmeister
 * @since Envoy Server Standalone v0.1-alpha
 */
public class UserStatusChangeProcessor implements ObjectProcessor<UserStatusChange> {

	private static ObjectWriteProxy		writeProxy;
	private static PersistenceManager	persistenceManager	= PersistenceManager.getInstance();

	private static final Logger logger = EnvoyLog.getLogger(UserStatusChangeProcessor.class);

	@Override
	public Class<UserStatusChange> getInputClass() { return UserStatusChange.class; }

	@Override
	public void process(UserStatusChange input, long socketID, ObjectWriteProxy writeProxy) throws IOException {
		// new status should not equal old status
		if (input.get().equals(persistenceManager.getUserByID(input.getID()).getStatus())) {
			logger.warning("Received an unnecessary UserStatusChange");
			return;
		}
		updateUserStatus(input);
	}

	/**
	 * Sets the {@link UserStatus} for a given user. Both offline contacts and
	 * currently online contacts are notified.
	 *
	 * @param user the {@link UserStatusChange} that signals the change
	 * @since Envoy Server Standalone v0.1-alpha
	 */

	public static void updateUserStatus(User user) {

		// Handling for newly logged in clients
		persistenceManager.updateContact(user);

		// Handling for contacts that are already online
		notifyContacts(user);
	}

	/**
	 * @param evt the {@link UserStatusChange}
	 * @since Envoy Server Standalone v0.1-alpha
	 */
	public static void updateUserStatus(UserStatusChange evt) { updateUserStatus(persistenceManager.getUserByID(evt.getID())); }

	/**
	 * notifies active contacts of this {@link User} that his {@link UserStatus} has
	 * changed
	 *
	 * @param user the {@link User}
	 * @since Envoy Server Standalone v0.1-alpha
	 */
	private static void notifyContacts(User user) {
		UserStatusChange	evt					= new UserStatusChange(user.getID(), user.getStatus());
		ConnectionManager		connectionManager	= ConnectionManager.getInstance();
		try {
			for (envoy.server.data.Contact contact : user.getContacts())
				if (connectionManager.isOnline(contact.getID())) writeProxy.write(connectionManager.getSocketID(contact.getID()), evt);
		} catch (IOException e) {
			e.printStackTrace();
			logger.warning("Could not notify online contacts of user " + evt.getID() + " that his status has been changed");
		}
	}

	/**
	 * This method is only called by the LoginCredentialProcessor because every
	 * user needs to login (open a socket) before changing his status.
	 * Needed to ensure propagation of events because an uninitialized writeProxy
	 * would cause problems.
	 *
	 * @param writeProxy the writeProxy that is used to send objects back to clients
	 * @since Envoy Server Standalone v0.1-alpha
	 */
	public static void setWriteProxy(ObjectWriteProxy writeProxy) { UserStatusChangeProcessor.writeProxy = writeProxy; }
	// TODO may cause an problem if two clients log in at the same time.
	// Change Needed.
}
